"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

from . import utils
from .errors import ClientException
from .utils import cached_slot_property
from .voice_client import VoiceClient

if TYPE_CHECKING:
    from . import abc
    from .abc import T as ConnectReturn
    from .channel import DMChannel, GroupChannel
    from .client import Client
    from .member import VoiceState
    from .message import Message
    from .state import ConnectionState
    from .user import BaseUser, User

    _PrivateChannel = Union[abc.DMChannel, abc.GroupChannel]

__all__ = (
    "CallMessage",
    "PrivateCall",
    "GroupCall",
)


def _running_only(func: Callable):
    def decorator(self: Call, *args, **kwargs):
        if self._ended:
            raise ClientException("Call is over")
        else:
            return func(self, *args, **kwargs)

    return decorator


class CallMessage:
    """Represents a group call message from Discord.

    This is only received in cases where the message type is equivalent to
    :attr:`MessageType.call`.

    Attributes
    -----------
    ended_timestamp: Optional[:class:`datetime.datetime`]
        An aware UTC datetime object that represents the time that the call has ended.
    participants: List[:class:`User`]
        A list of users that participated in the call.
    message: :class:`Message`
        The message associated with this call message.
    """

    __slots__ = ("message", "ended_timestamp", "participants")

    def __init__(
        self,
        message: Message,
        *,
        participants: List[User],
        ended_timestamp: Optional[str],
    ) -> None:
        self.message = message
        self.ended_timestamp = utils.parse_time(ended_timestamp)
        self.participants = participants

    @property
    def call_ended(self) -> bool:
        """:class:`bool`: Indicates if the call has ended."""
        return self.ended_timestamp is not None

    @property
    def initiator(self) -> User:
        """:class:`.abc.User`: Returns the user that started the call."""
        return self.message.author  # type: ignore # Cannot be a Member in private messages

    @property
    def channel(self) -> _PrivateChannel:
        """:class:`.abc.PrivateChannel`: The private channel associated with this message."""
        return self.message.channel  # type: ignore # Can only be a private channel here

    @property
    def duration(self) -> datetime.timedelta:
        """Queries the duration of the call.

        If the call has not ended then the current duration will
        be returned.

        Returns
        ---------
        :class:`datetime.timedelta`
            The timedelta object representing the duration.
        """
        if self.ended_timestamp is None:
            return utils.utcnow() - self.message.created_at
        else:
            return self.ended_timestamp - self.message.created_at


class PrivateCall:
    """Represents the actual group call from Discord.

    This is accompanied with a :class:`CallMessage` denoting the information.

    .. versionadded:: 1.9

    Attributes
    -----------
    channel: :class:`DMChannel`
        The channel the call is in.
    unavailable: :class:`bool`
        Denotes if this call is unavailable.
    region: :class:`str`
        The region the call is being hosted at.

        .. versionchanged:: 2.0
            The type of this attribute has changed to :class:`str`.
    """

    __slots__ = (
        "_state",
        "_ended",
        "channel",
        "_cs_message",
        "_ringing",
        "_message_id",
        "region",
        "unavailable",
    )

    if TYPE_CHECKING:
        channel: DMChannel

    def __init__(
        self,
        *,
        data: dict,
        state: ConnectionState,
        message: Optional[Message],
        channel: abc.PrivateChannel,
    ) -> None:
        self._state = state
        self._cs_message = message
        self.channel = channel  # type: ignore # Will always be a DMChannel here
        self._ended: bool = False

        self._update(data)

    def _delete(self) -> None:
        self._ringing = tuple()
        self._ended = True

    def _get_recipients(self) -> Tuple[BaseUser, ...]:
        channel = self.channel
        return channel.me, channel.recipient

    def _is_participating(self, user: BaseUser) -> bool:
        state = self.voice_state_for(user)
        return bool(state and state.channel and state.channel.id == self.channel.id)

    def _update(self, data) -> None:
        self._message_id = int(data["message_id"])
        self.unavailable = data.get("unavailable", False)
        try:
            self.region: str = data["region"]
        except KeyError:
            pass

        channel = self.channel
        recipients = self._get_recipients()
        lookup = {u.id: u for u in recipients}
        self._ringing = tuple(filter(None, map(lookup.get, data.get("ringing", []))))

        for vs in data.get("voice_states", []):
            self._state._update_voice_state(vs, channel.id)

    @property
    def ringing(self) -> List[BaseUser]:
        """List[:class:`.abc.User`]: A list of users that are currently being rung to join the call."""
        return list(self._ringing)

    @property
    def initiator(self) -> Optional[User]:
        """Optional[:class:`.abc.User`]: Returns the user that started the call. Returns ``None`` if the message is not cached."""
        return getattr(self.message, "author", None)

    @property
    def connected(self) -> bool:
        """:class:`bool`: Returns whether you're in the call (this does not mean you're in the call through the library)."""
        return self._is_participating(self.channel.me)

    @property
    def members(self) -> List[BaseUser]:
        """List[:class:`.abc.User`]: Returns all users that are currently in this call."""
        recipients = self._get_recipients()
        return [u for u in recipients if self._is_participating(u)]

    @property
    def voice_states(self) -> Dict[int, VoiceState]:
        """Mapping[:class:`int`, :class:`VoiceState`]: Returns a mapping of user IDs who have voice states in this call."""
        return {
            k: v
            for k, v in self._state._voice_states.items()
            if bool(v and v.channel and v.channel.id == self.channel.id)
        }

    @cached_slot_property("_cs_message")
    def message(self) -> Optional[Message]:
        """Optional[:class:`Message`]: The message associated with this call. Sometimes may not be cached."""
        return self._state._get_message(self._message_id)

    async def fetch_message(self) -> Message:
        """|coro|

        Fetches and caches the message associated with this call.

        Raises
        -------
        HTTPException
            Retrieving the message failed.

        Returns
        -------
        :class:`Message`
            The message associated with this call.
        """
        message = await self.channel.fetch_message(self._message_id)
        state = self._state
        if self.message is None:
            if state._messages is not None:
                state._messages.append(message)
            self._cs_message = message
        return message

    async def change_region(self, region: str) -> None:
        """|coro|

        Changes the channel's voice region.

        Parameters
        -----------
        region: :class:`str`
            A region to change the voice region to.

            .. versionchanged:: 2.0
                The type of this parameter has changed to :class:`str`.

        Raises
        -------
        HTTPException
            Failed to change the channel's voice region.
        """
        await self._state.http.change_call_voice_region(self.channel.id, region)

    @_running_only
    async def ring(self) -> None:
        """|coro|

        Rings the other recipient.

        Raises
        -------
        Forbidden
            Not allowed to ring the other recipient.
        HTTPException
            Ringing failed.
        ClientException
            The call has ended.
        """
        channel = self.channel
        await self._state.http.ring(channel.id)

    @_running_only
    async def stop_ringing(self) -> None:
        """|coro|

        Stops ringing the other recipient.

        Raises
        -------
        HTTPException
            Stopping the ringing failed.
        ClientException
            The call has ended.
        """
        channel = self.channel
        await self._state.http.stop_ringing(channel.id, channel.recipient.id)

    @_running_only
    async def connect(
        self,
        *,
        timeout: float = 60.0,
        reconnect: bool = True,
        cls: Callable[[Client, abc.VocalChannel], ConnectReturn] = VoiceClient,
    ) -> ConnectReturn:
        """|coro|

        Connects to voice and creates a :class:`~discord.VoiceClient` to establish
        your connection to the voice server.

        There is an alias of this called :attr:`join`.

        Parameters
        -----------
        timeout: :class:`float`
            The timeout in seconds to wait for the voice endpoint.
        reconnect: :class:`bool`
            Whether the bot should automatically attempt
            a reconnect if a part of the handshake fails
            or the gateway goes down.
        cls: Type[:class:`~discord.VoiceProtocol`]
            A type that subclasses :class:`~discord.VoiceProtocol` to connect with.
            Defaults to :class:`~discord.VoiceClient`.

        Raises
        -------
        asyncio.TimeoutError
            Could not connect to the voice channel in time.
        ~discord.ClientException
            You are already connected to a voice channel.
        ~discord.opus.OpusNotLoaded
            The opus library has not been loaded.

        Returns
        --------
        :class:`~discord.VoiceProtocol`
            A voice client that is fully connected to the voice server.
        """
        return await self.channel.connect(
            timeout=timeout, reconnect=reconnect, cls=cls, ring=False
        )

    @_running_only
    async def join(
        self,
        *,
        timeout: float = 60.0,
        reconnect: bool = True,
        cls: Callable[[Client, abc.VocalChannel], ConnectReturn] = VoiceClient,
    ) -> ConnectReturn:
        """|coro|

        Connects to voice and creates a :class:`~discord.VoiceClient` to establish
        your connection to the voice server.

        This is an alias of :attr:`connect`.

        Parameters
        -----------
        timeout: :class:`float`
            The timeout in seconds to wait for the voice endpoint.
        reconnect: :class:`bool`
            Whether the bot should automatically attempt
            a reconnect if a part of the handshake fails
            or the gateway goes down.
        cls: Type[:class:`~discord.VoiceProtocol`]
            A type that subclasses :class:`~discord.VoiceProtocol` to connect with.
            Defaults to :class:`~discord.VoiceClient`.

        Raises
        -------
        asyncio.TimeoutError
            Could not connect to the voice channel in time.
        ~discord.ClientException
            You are already connected to a voice channel.
        ~discord.opus.OpusNotLoaded
            The opus library has not been loaded.

        Returns
        --------
        :class:`~discord.VoiceProtocol`
            A voice client that is fully connected to the voice server.
        """
        return await self.connect(timeout=timeout, reconnect=reconnect, cls=cls)

    @_running_only
    async def disconnect(self, force: bool = False) -> None:
        """|coro|

        Disconnects this voice client from voice.

        There is an alias of this called :attr:`leave`.
        """
        state = self._state
        if not (client := state._get_voice_client(self.channel.me.id)):
            return

        return await client.disconnect(force=force)

    @_running_only
    async def leave(self, force: bool = False) -> None:
        """|coro|

        Disconnects this voice client from voice.

        This is an alias of :attr:`disconnect`.
        """
        return await self.disconnect(force=force)

    def voice_state_for(self, user: abc.Snowflake) -> Optional[VoiceState]:
        """Retrieves the :class:`VoiceState` for a specified :class:`User`.

        If the :class:`User` has no voice state then this function returns
        ``None``.

        Parameters
        ------------
        user: :class:`User`
            The user to retrieve the voice state for.

        Returns
        --------
        Optional[:class:`VoiceState`]
            The voice state associated with this user.
        """
        return self._state._voice_state_for(user.id)


class GroupCall(PrivateCall):
    """Represents a Discord group call.

    This is accompanied with a :class:`CallMessage` denoting the information.

    Attributes
    -----------
    channel: :class:`GroupChannel`
        The channel the group call is in.
    unavailable: :class:`bool`
        Denotes if this group call is unavailable.
    region: :class:`str`
        The region the group call is being hosted in.

        .. versionchanged:: 2.0
            The type of this attribute has changed to :class:`str`.
    """

    __slots__ = ()

    if TYPE_CHECKING:
        channel: GroupChannel

    def _get_recipients(self) -> Tuple[BaseUser, ...]:
        channel = self.channel
        return *channel.recipients, channel.me

    @_running_only
    async def ring(self, *recipients: abc.Snowflake) -> None:
        r"""|coro|

        Rings the specified recipients.

        Parameters
        -----------
        \*recipients: :class:`User`
            The recipients to ring. The default is to ring all recipients.

        Raises
        -------
        HTTPException
            Stopping the ringing failed.
        ClientException
            The call has ended.
        """
        await self._state.http.ring(self.channel.id, *{r.id for r in recipients})

    @_running_only
    async def stop_ringing(self, *recipients: abc.Snowflake) -> None:
        r"""|coro|

        Stops ringing the specified recipients.

        Parameters
        -----------
        \*recipients: :class:`User`
            The recipients to stop ringing.

        Raises
        -------
        HTTPException
            Ringing failed.
        ClientException
            The call has ended.
        """
        channel = self.channel
        await self._state.http.stop_ringing(
            channel.id, *{r.id for r in recipients or channel.recipients}
        )


Call = Union[PrivateCall, GroupCall]



from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.welcome_screen import (
    WelcomeScreen,
    WelcomeScreenChannel,
)

__all__ = ["Snowflake", "WelcomeScreen", "WelcomeScreenChannel"]



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import json
from base64 import b64encode
from random import choice
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

from .utils import MISSING

if TYPE_CHECKING:
    from typing_extensions import Self

    from .enums import ChannelType
    from .types.snowflake import Snowflake

# fmt: off
__all__ = (
    'ContextProperties',
)
# fmt: on


class ContextPropertiesMeta(type):
    if TYPE_CHECKING:

        def __getattribute__(self, name: str) -> Callable[[], Self]: ...

    def __new__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]):
        cls = super().__new__(cls, name, bases, attrs)
        locations = attrs.get("LOCATIONS", {})
        sources = attrs.get("SOURCES", {})

        def build_location(location: str) -> classmethod:
            def f(cls) -> Self:
                data = {"location": location}
                return cls(data)

            return classmethod(f)

        def build_source(source: str) -> classmethod:
            def f(cls) -> Self:
                data = {"source": source}
                return cls(data)

            return classmethod(f)

        for location in locations:
            if location:
                setattr(
                    cls,
                    f"from_{location.lower().replace(' ', '_').replace('/', '')}",
                    build_location(location),
                )

        for source in sources:
            if source:
                setattr(
                    cls,
                    f"from_{source.lower().replace(' ', '_')}",
                    build_source(source),
                )

        return cls


class ContextProperties(metaclass=ContextPropertiesMeta):
    """Represents the Discord X-Context-Properties header.

    This header is essential for certain actions (e.g. joining guilds, friend requesting).
    """

    __slots__ = ("_data",)

    LOCATIONS = {
        None: "e30=",
        "Friends": "eyJsb2NhdGlvbiI6IkZyaWVuZHMifQ==",
        "ContextMenu": "eyJsb2NhdGlvbiI6IkNvbnRleHRNZW51In0=",
        "Context Menu": "eyJsb2NhdGlvbiI6IkNvbnRleHQgTWVudSJ9",
        "User Profile": "eyJsb2NhdGlvbiI6IlVzZXIgUHJvZmlsZSJ9",
        "Add Friend": "eyJsb2NhdGlvbiI6IkFkZCBGcmllbmQifQ==",
        "Guild Header": "eyJsb2NhdGlvbiI6Ikd1aWxkIEhlYWRlciJ9",
        "Group DM": "eyJsb2NhdGlvbiI6Ikdyb3VwIERNIn0=",
        "DM Channel": "eyJsb2NhdGlvbiI6IkRNIENoYW5uZWwifQ==",
        "/app": "eyJsb2NhdGlvbiI6ICIvYXBwIn0=",
        "Login": "eyJsb2NhdGlvbiI6IkxvZ2luIn0=",
        "Register": "eyJsb2NhdGlvbiI6IlJlZ2lzdGVyIn0=",
        "Verify Email": "eyJsb2NhdGlvbiI6IlZlcmlmeSBFbWFpbCJ9",
        "New Group DM": "eyJsb2NhdGlvbiI6Ik5ldyBHcm91cCBETSJ9",
        "Add Friends to DM": "eyJsb2NhdGlvbiI6IkFkZCBGcmllbmRzIHRvIERNIn0=",
        "Group DM Invite Create": "eyJsb2NhdGlvbiI6Ikdyb3VwIERNIEludml0ZSBDcmVhdGUifQ==",
        "Stage Channel": "eyJsb2NhdGlvbiI6IlN0YWdlIENoYW5uZWwifQ==",
    }

    SOURCES = {
        None: "e30=",
        "Chat Input Blocker - Lurker Mode": "eyJzb3VyY2UiOiJDaGF0IElucHV0IEJsb2NrZXIgLSBMdXJrZXIgTW9kZSJ9",
        "Notice - Lurker Mode": "eyJzb3VyY2UiOiJOb3RpY2UgLSBMdXJrZXIgTW9kZSJ9",
    }

    def __init__(self, data: dict) -> None:
        self._data: Dict[str, Snowflake] = data

    def _encode_data(self) -> str:
        try:
            target = self.target
            return self.LOCATIONS.get(target, self.SOURCES[target])
        except KeyError:
            return b64encode(
                json.dumps(self._data, separators=(",", ":")).encode()
            ).decode("utf-8")

    @classmethod
    def empty(cls) -> Self:
        return cls({})

    @classmethod
    def from_accept_invite_page(
        cls,
        *,
        guild_id: Optional[Snowflake] = None,
        channel_id: Optional[Snowflake] = None,
        channel_type: Optional[ChannelType] = None,
    ) -> Self:
        data: Dict[str, Snowflake] = {
            "location": "Accept Invite Page",
        }
        if guild_id:
            data["location_guild_id"] = str(guild_id)
        if channel_id:
            data["location_channel_id"] = str(channel_id)
        if channel_type:
            data["location_channel_type"] = int(channel_type)
        return cls(data)

    @classmethod
    def from_join_guild(
        cls,
        *,
        guild_id: Snowflake = MISSING,
        channel_id: Snowflake = MISSING,
        channel_type: ChannelType = MISSING,
    ) -> Self:
        data: Dict[str, Snowflake] = {
            "location": "Join Guild",
        }
        if guild_id is not MISSING:
            data["location_guild_id"] = str(guild_id)
        if channel_id is not MISSING:
            data["location_channel_id"] = str(channel_id)
        if channel_type is not MISSING:
            data["location_channel_type"] = int(channel_type)
        return cls(data)

    @classmethod
    def from_invite_button_embed(
        cls,
        *,
        guild_id: Optional[Snowflake],
        channel_id: Snowflake,
        message_id: Snowflake,
        channel_type: Optional[ChannelType],
    ) -> Self:
        data = {
            "location": "Invite Button Embed",
            "location_guild_id": str(guild_id) if guild_id else None,
            "location_channel_id": str(channel_id),
            "location_channel_type": int(channel_type) if channel_type else None,
            "location_message_id": str(message_id),
        }
        return cls(data)

    @classmethod
    def from_lurking(cls, source: str = MISSING) -> Self:
        data = {
            "source": source
            or choice(("Chat Input Blocker - Lurker Mode", "Notice - Lurker Mode"))
        }
        return cls(data)

    @property
    def target(self) -> Optional[str]:
        return self._data.get("location", self._data.get("source"))  # type: ignore

    @property
    def value(self) -> str:
        return self._encode_data()

    def __str__(self) -> str:
        return self.target or "None"

    def __repr__(self) -> str:
        return f"<ContextProperties target={self.target!r}>"

    def __eq__(self, other) -> bool:
        return isinstance(other, ContextProperties) and self.value == other.value

    def __ne__(self, other) -> bool:
        if isinstance(other, ContextProperties):
            return self.value != other.value
        return True



from discord_self._vendor.discord.types.integration import ConnectionIntegration
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import (
    APIUser,
    Connection,
    ConnectionAccessToken,
    ConnectionAuthorization,
    ConnectionType,
    ConnectionVisibilty,
    GuildAffinities,
    GuildAffinity,
    Note,
    PartialConnection,
    PartialUser,
    PremiumType,
    ProtoSettings,
    ProtoSettingsType,
    Relationship,
    RelationshipType,
    User,
    UserAffinities,
    UserAffinity,
)

__all__ = [
    "APIUser",
    "Connection",
    "ConnectionAccessToken",
    "ConnectionAuthorization",
    "ConnectionIntegration",
    "ConnectionType",
    "ConnectionVisibilty",
    "GuildAffinities",
    "GuildAffinity",
    "Note",
    "PartialConnection",
    "PartialUser",
    "PremiumType",
    "ProtoSettings",
    "ProtoSettingsType",
    "Relationship",
    "RelationshipType",
    "Snowflake",
    "User",
    "UserAffinities",
    "UserAffinity",
]



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from .billing import PaymentSource
from .enums import (
    PaymentGateway,
    PaymentStatus,
    SubscriptionType,
    try_enum,
)
from .flags import PaymentFlags
from .mixins import Hashable
from .store import SKU
from .subscriptions import Subscription
from .utils import _get_as_snowflake, parse_time

if TYPE_CHECKING:
    from .entitlements import Entitlement
    from .state import ConnectionState
    from .types.payments import (
        PartialPayment as PartialPaymentPayload,
    )
    from .types.payments import (
        Payment as PaymentPayload,
    )

__all__ = (
    "Payment",
    "EntitlementPayment",
)


class Payment(Hashable):
    """Represents a payment to Discord.

    .. container:: operations

        .. describe:: x == y

            Checks if two payments are equal.

        .. describe:: x != y

            Checks if two payments are not equal.

        .. describe:: hash(x)

            Returns the payment's hash.

        .. describe:: str(x)

            Returns the payment's description.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the payment.
    amount: :class:`int`
        The amount of the payment.
    amount_refunded: :class:`int`
        The amount refunded from the payment, if any.
    tax: :class:`int`
        The amount of tax paid.
    tax_inclusive: :class:`bool`
        Whether the amount is inclusive of all taxes.
    currency: :class:`str`
        The currency the payment was made in.
    description: :class:`str`
        What the payment was for.
    status: :class:`PaymentStatus`
        The status of the payment.
    created_at: :class:`datetime.datetime`
        The time the payment was made.
    sku: Optional[:class:`SKU`]
        The SKU the payment was for, if applicable.
    sku_id: Optional[:class:`int`]
        The ID of the SKU the payment was for, if applicable.
    sku_price: Optional[:class:`int`]
        The price of the SKU the payment was for, if applicable.
    subscription_plan_id: Optional[:class:`int`]
        The ID of the subscription plan the payment was for, if applicable.
    subscription: Optional[:class:`Subscription`]
        The subscription the payment was for, if applicable.
    payment_source: Optional[:class:`PaymentSource`]
        The payment source the payment was made with.
    payment_gateway: Optional[:class:`PaymentGateway`]
        The payment gateway the payment was made with, if applicable.
    payment_gateway_payment_id: Optional[:class:`str`]
        The ID of the payment on the payment gateway, if any.
    invoice_url: Optional[:class:`str`]
        The URL to download the VAT invoice for this payment, if available.
    refund_invoices_urls: List[:class:`str`]
        A list of URLs to download VAT credit notices for refunds on this payment, if available.
    refund_disqualification_reasons: List[:class:`str`]
        A list of reasons why the payment cannot be refunded, if any.
    """

    __slots__ = (
        "id",
        "amount",
        "amount_refunded",
        "tax",
        "tax_inclusive",
        "currency",
        "description",
        "status",
        "created_at",
        "sku",
        "sku_id",
        "sku_price",
        "subscription_plan_id",
        "subscription",
        "payment_source",
        "payment_gateway",
        "payment_gateway_payment_id",
        "invoice_url",
        "refund_invoices_urls",
        "refund_disqualification_reasons",
        "_flags",
        "_state",
    )

    def __init__(self, *, data: PaymentPayload, state: ConnectionState):
        self._state: ConnectionState = state
        self._update(data)

    def _update(self, data: PaymentPayload) -> None:
        state = self._state

        self.id: int = int(data["id"])
        self.amount: int = data["amount"]
        self.amount_refunded: int = data.get("amount_refunded") or 0
        self.tax: int = data.get("tax") or 0
        self.tax_inclusive: bool = data.get("tax_inclusive", True)
        self.currency: str = data.get("currency", "usd")
        self.description: str = data["description"]
        self.status: PaymentStatus = try_enum(PaymentStatus, data["status"])
        self.created_at: datetime = parse_time(data["created_at"])
        self.sku: Optional[SKU] = (
            SKU(data=data["sku"], state=state) if "sku" in data else None
        )
        self.sku_id: Optional[int] = _get_as_snowflake(data, "sku_id")
        self.sku_price: Optional[int] = data.get("sku_price")
        self.subscription_plan_id: Optional[int] = _get_as_snowflake(
            data, "sku_subscription_plan_id"
        )
        self.payment_gateway: Optional[PaymentGateway] = (
            try_enum(PaymentGateway, data["payment_gateway"])
            if "payment_gateway" in data
            else None
        )
        self.payment_gateway_payment_id: Optional[str] = data.get(
            "payment_gateway_payment_id"
        )
        self.invoice_url: Optional[str] = data.get("downloadable_invoice")
        self.refund_invoices_urls: List[str] = data.get(
            "downloadable_refund_invoices", []
        )
        self.refund_disqualification_reasons: List[str] = data.get(
            "premium_refund_disqualification_reasons", []
        )
        self._flags: int = data.get("flags", 0)

        # The subscription object does not include the payment source ID
        self.payment_source: Optional[PaymentSource] = (
            PaymentSource(data=data["payment_source"], state=state)
            if "payment_source" in data
            else None
        )
        if "subscription" in data and self.payment_source:
            data["subscription"]["payment_source_id"] = self.payment_source.id  # type: ignore
        self.subscription: Optional[Subscription] = (
            Subscription(data=data["subscription"], state=state)
            if "subscription" in data
            else None
        )

    def __repr__(self) -> str:
        return f"<Payment id={self.id} amount={self.amount} currency={self.currency} status={self.status}>"

    def __str__(self) -> str:
        return self.description

    def is_subscription(self) -> bool:
        """:class:`bool`: Whether the payment was for a subscription."""
        return self.subscription is not None

    def is_premium_subscription(self) -> bool:
        """:class:`bool`: Whether the payment was for a Discord premium subscription."""
        return (
            self.subscription is not None
            and self.subscription.type == SubscriptionType.premium
        )

    def is_premium_subscription_gift(self) -> bool:
        """:class:`bool`: Whether the payment was for a Discord premium subscription gift."""
        return (
            self.flags.gift
            and self.sku_id in self._state.premium_subscriptions_sku_ids.values()
        )

    def is_purchased_externally(self) -> bool:
        """:class:`bool`: Whether the payment was made externally."""
        return self.payment_gateway in (PaymentGateway.apple, PaymentGateway.google)

    @property
    def flags(self) -> PaymentFlags:
        """:class:`PaymentFlags`: Returns the payment's flags."""
        return PaymentFlags._from_value(self._flags)

    async def void(self) -> None:
        """|coro|

        Void the payment. Only applicable for payments of status :attr:`PaymentStatus.pending`.

        Raises
        ------
        HTTPException
            Voiding the payment failed.
        """
        await self._state.http.void_payment(self.id)
        self.status = PaymentStatus.failed

    async def refund(self, reason: Optional[int] = None) -> None:
        """|coro|

        Refund the payment.

        Raises
        ------
        HTTPException
            Refunding the payment failed.
        """
        # reason here is an enum (0-8), but I was unable to find the enum values
        # Either way, it's optional and this endpoint isn't really used anyway
        await self._state.http.refund_payment(self.id, reason)
        self.status = PaymentStatus.refunded


class EntitlementPayment(Hashable):
    """Represents a partial payment for an entitlement.

    .. container:: operations

        .. describe:: x == y

            Checks if two payments are equal.

        .. describe:: x != y

            Checks if two payments are not equal.

        .. describe:: hash(x)

            Returns the payment's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    entitlement: :class:`Entitlement`
        The entitlement the payment is for.
    id: :class:`int`
        The ID of the payment.
    amount: :class:`int`
        The amount of the payment.
    tax: :class:`int`
        The amount of tax paid.
    tax_inclusive: :class:`bool`
        Whether the amount is inclusive of all taxes.
    currency: :class:`str`
        The currency the payment was made in.
    """

    __slots__ = ("entitlement", "id", "amount", "tax", "tax_inclusive", "currency")

    def __init__(self, *, data: PartialPaymentPayload, entitlement: Entitlement):
        self.entitlement = entitlement
        self.id: int = int(data["id"])
        self.amount: int = data["amount"]
        self.tax: int = data.get("tax") or 0
        self.tax_inclusive: bool = data.get("tax_inclusive", True)
        self.currency: str = data.get("currency", "usd")

    def __repr__(self) -> str:
        return f"<EntitlementPayment id={self.id} amount={self.amount} currency={self.currency}>"



from discord_self._vendor.discord.activity import (
    Activity,
    BaseActivity,
    CustomActivity,
    Game,
    Session,
    Spotify,
    Streaming,
)

__all__ = [
    "Activity",
    "BaseActivity",
    "CustomActivity",
    "Game",
    "Session",
    "Spotify",
    "Streaming",
]



from discord_self._vendor.discord.player import (
    AudioSource,
    FFmpegAudio,
    FFmpegOpusAudio,
    FFmpegPCMAudio,
    PCMAudio,
    PCMVolumeTransformer,
)

__all__ = [
    "AudioSource",
    "FFmpegAudio",
    "FFmpegOpusAudio",
    "FFmpegPCMAudio",
    "PCMAudio",
    "PCMVolumeTransformer",
]



from discord_self._vendor.discord.types.entitlements import (
    Entitlement,
    GatewayGift,
    Gift,
    GiftBatch,
)
from discord_self._vendor.discord.types.payments import PartialPayment
from discord_self._vendor.discord.types.promotions import Promotion
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.store import SKU, StoreListing
from discord_self._vendor.discord.types.subscriptions import (
    PartialSubscriptionPlan,
    SubscriptionPlan,
    SubscriptionTrial,
)
from discord_self._vendor.discord.types.user import PartialUser

__all__ = [
    "Entitlement",
    "GatewayGift",
    "Gift",
    "GiftBatch",
    "PartialPayment",
    "PartialSubscriptionPlan",
    "PartialUser",
    "Promotion",
    "SKU",
    "Snowflake",
    "StoreListing",
    "SubscriptionPlan",
    "SubscriptionTrial",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, AsyncIterator, List, Optional, Union, overload

from . import utils
from .asset import Asset
from .enums import (
    ApplicationMembershipState,
    PayoutAccountStatus,
    PayoutReportType,
    PayoutStatus,
    try_enum,
)
from .metadata import Metadata
from .mixins import Hashable
from .object import Object
from .user import User, _UserTag

if TYPE_CHECKING:
    from datetime import date

    from .abc import Snowflake, SnowflakeTime
    from .application import Application, Company
    from .state import ConnectionState
    from .types.team import Team as TeamPayload
    from .types.team import TeamMember as TeamMemberPayload
    from .types.team import TeamPayout as TeamPayoutPayload
    from .types.user import PartialUser as PartialUserPayload

__all__ = (
    "Team",
    "TeamMember",
    "TeamPayout",
)

MISSING = utils.MISSING


class Team(Hashable):
    """Represents an application team.

    .. container:: operations

        .. describe:: x == y

            Checks if two teams are equal.

        .. describe:: x != y

            Checks if two teams are not equal.

        .. describe:: hash(x)

            Return the team's hash.

        .. describe:: str(x)

            Returns the team's name.

    .. versionadded:: 2.0

    Attributes
    -------------
    id: :class:`int`
        The team ID.
    name: :class:`str`
        The team name.
    owner_id: :class:`int`
        The team's owner ID.
    members: List[:class:`TeamMember`]
        The team's members.

        .. note::

            In almost all cases, a call to :meth:`fetch_members`
            is required to populate this list past (sometimes) the owner.
    payout_account_status: Optional[:class:`PayoutAccountStatus`]
        The team's payout account status, if any and available.
    stripe_connect_account_id: Optional[:class:`str`]
        The account ID representing the Stripe Connect account the
        team's payout account is linked to, if any and available.
    """

    __slots__ = (
        "_state",
        "id",
        "name",
        "_icon",
        "owner_id",
        "members",
        "payout_account_status",
        "stripe_connect_account_id",
    )

    def __init__(self, state: ConnectionState, data: TeamPayload):
        self._state: ConnectionState = state

        self.members: List[TeamMember] = []
        self.payout_account_status: Optional[PayoutAccountStatus] = None
        self.stripe_connect_account_id: Optional[str] = None
        self._update(data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name}>"

    def __str__(self) -> str:
        return self.name

    def _update(self, data: TeamPayload):
        state = self._state

        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self._icon: Optional[str] = data["icon"]
        self.owner_id = owner_id = int(data["owner_user_id"])

        if "members" in data:
            self.members = [
                TeamMember(self, state=state, data=member)
                for member in data.get("members", [])
            ]

        if not self.owner:
            owner = self._state.get_user(owner_id)
            if owner:
                user: PartialUserPayload = owner._to_minimal_user_json()
                member: TeamMemberPayload = {
                    "user": user,
                    "team_id": self.id,
                    "membership_state": 2,
                    "permissions": ["*"],
                }
                self.members.append(TeamMember(self, self._state, member))

        if "payout_account_status" in data:
            self.payout_account_status = try_enum(
                PayoutAccountStatus, data.get("payout_account_status")
            )
        if "stripe_connect_account_id" in data:
            self.stripe_connect_account_id = data.get("stripe_connect_account_id")

    @property
    def icon(self) -> Optional[Asset]:
        """Optional[:class:`.Asset`]: Retrieves the team's icon asset, if any."""
        if self._icon is None:
            return None
        return Asset._from_icon(self._state, self.id, self._icon, path="team")

    @property
    def default_icon(self) -> Asset:
        """:class:`Asset`: Returns the default icon for the team. This is calculated by the team's ID."""
        return Asset._from_default_avatar(self._state, int(self.id) % 5)

    @property
    def display_icon(self) -> Asset:
        """:class:`Asset`: Returns the team's display icon.

        For regular teams this is just their default icon or uploaded icon.
        """
        return self.icon or self.default_icon

    @property
    def owner(self) -> Optional[TeamMember]:
        """Optional[:class:`TeamMember`]: The team's owner, if available."""
        return utils.get(self.members, id=self.owner_id)

    async def edit(
        self,
        *,
        name: str = MISSING,
        icon: Optional[bytes] = MISSING,
        owner: Snowflake = MISSING,
    ) -> None:
        """|coro|

        Edits the team.

        All parameters are optional.

        Parameters
        -----------
        name: :class:`str`
            The name of the team.
        icon: Optional[:class:`bytes`]
            The icon of the team.
        owner: :class:`User`
            The team's owner.

        Raises
        -------
        Forbidden
            You do not have permissions to edit the team.
        HTTPException
            Editing the team failed.
        """
        payload = {}
        if name is not MISSING:
            payload["name"] = name
        if icon is not MISSING:
            if icon is not None:
                payload["icon"] = utils._bytes_to_base64_data(icon)
            else:
                payload["icon"] = ""
        if owner is not MISSING:
            payload["owner_user_id"] = owner.id

        data = await self._state.http.edit_team(self.id, payload)
        self._update(data)

    async def applications(self) -> List[Application]:
        """|coro|

        Retrieves the team's applications.

        Returns
        --------
        List[:class:`TeamMember`]
            The team's applications.

        Raises
        -------
        Forbidden
            You do not have permissions to fetch the team's applications.
        HTTPException
            Retrieving the team applications failed.
        """
        from .application import Application

        state = self._state
        data = await state.http.get_team_applications(self.id)
        return [Application(state=state, data=app, team=self) for app in data]

    async def fetch_members(self) -> List[TeamMember]:
        """|coro|

        Retrieves and caches the team's members.

        Returns
        --------
        List[:class:`TeamMember`]
            The team's members.

        Raises
        -------
        Forbidden
            You do not have permissions to fetch the team's members.
        HTTPException
            Retrieving the team members failed.
        """
        data = await self._state.http.get_team_members(self.id)
        members = [TeamMember(self, self._state, member) for member in data]
        self.members = members
        return members

    @overload
    async def invite_member(self, user: _UserTag, /) -> TeamMember: ...

    @overload
    async def invite_member(self, user: str, /) -> TeamMember: ...

    @overload
    async def invite_member(
        self, username: str, discriminator: str, /
    ) -> TeamMember: ...

    async def invite_member(self, *args: Union[_UserTag, str]) -> TeamMember:
        """|coro|

        Invites a member to the team.

        This function can be used in multiple ways.

        .. code-block:: python

            # Passing a user object:
            await team.invite_member(user)

            # Passing a stringified user:
            await team.invite_member('Jake#0001')

            # Passing a username and discriminator:
            await team.invite_member('Jake', '0001')

        Parameters
        -----------
        user: Union[:class:`User`, :class:`str`]
            The user to invite.
        username: :class:`str`
            The username of the user to invite.
        discriminator: :class:`str`
            The discriminator of the user to invite.

        Raises
        -------
        Forbidden
            You do not have permissions to invite the user.
        HTTPException
            Inviting the user failed.
        TypeError
            More than 2 parameters or less than 1 parameter were passed.

        Returns
        -------
        :class:`TeamMember`
            The new member.
        """
        username: str
        discrim: str
        if len(args) == 1:
            user = args[0]
            if isinstance(user, _UserTag):
                user = str(user)
            username, discrim = user.split("#")
        elif len(args) == 2:
            username, discrim = args  # type: ignore
        else:
            raise TypeError(
                f"invite_member() takes 1 or 2 arguments but {len(args)} were given"
            )

        state = self._state
        data = await state.http.invite_team_member(self.id, username, discrim)
        member = TeamMember(self, state, data)
        self.members.append(member)
        return member

    async def create_company(self, name: str, /) -> Company:
        """|coro|

        Creates a company for the team.

        Parameters
        -----------
        name: :class:`str`
            The name of the company.

        Raises
        -------
        Forbidden
            You do not have permissions to create a company.
        HTTPException
            Creating the company failed.

        Returns
        -------
        :class:`.Company`
            The created company.
        """
        from .application import Company

        state = self._state
        data = await state.http.create_team_company(self.id, name)
        return Company(data=data)

    async def payouts(
        self,
        *,
        limit: Optional[int] = 96,
        before: Optional[SnowflakeTime] = None,
    ) -> AsyncIterator[TeamPayout]:
        """Returns an :term:`asynchronous iterator` that enables receiving your team payouts.

        .. versionadded:: 2.0

        Examples
        ---------

        Usage ::

            total = 0
            async for payout in team.payouts():
                if payout.period_end:
                    total += payout.amount

        Flattening into a list: ::

            payments = [payout async for payout in team.payouts(limit=123)]
            # payments is now a list of TeamPayout...

        All parameters are optional.

        Parameters
        -----------
        limit: Optional[:class:`int`]
            The number of payouts to retrieve.
            If ``None``, retrieves every payout you have. Note, however,
            that this would make it a slow operation.
        before: Optional[Union[:class:`~discord.abc.Snowflake`, :class:`datetime.datetime`]]
            Retrieve payments before this date or payout.
            If a datetime is provided, it is recommended to use a UTC aware datetime.
            If the datetime is naive, it is assumed to be local time.

        Raises
        ------
        HTTPException
            The request to get team payouts failed.

        Yields
        -------
        :class:`~discord.TeamPayout`
            The payout received.
        """

        async def strategy(
            retrieve: int, before: Optional[Snowflake], limit: Optional[int]
        ):
            before_id = before.id if before else None
            data = await self._state.http.get_team_payouts(
                self.id, limit=retrieve, before=before_id
            )

            if data:
                if limit is not None:
                    limit -= len(data)

                before = Object(id=int(data[-1]["id"]))

            return data, before, limit

        if isinstance(before, datetime):
            before = Object(id=utils.time_snowflake(before, high=False))

        while True:
            retrieve = min(96 if limit is None else limit, 100)
            if retrieve < 1:
                return

            data, before, limit = await strategy(retrieve, before, limit)

            # Terminate loop on next iteration; there's no data left after this
            if len(data) < 96:
                limit = 0

            for payout in data:
                yield TeamPayout(data=payout, team=self)

    async def leave(self) -> None:
        """|coro|

        Leaves the team.

        .. note::

            You cannot leave a team that you own, you must delete it instead
            via :meth:`delete`.

        Raises
        -------
        Forbidden
            You do not have permissions to leave the team.
        HTTPException
            Leaving the team failed.
        """
        await self._state.http.remove_team_member(self.id, self._state.self_id)  # type: ignore

    async def delete(self) -> None:
        """|coro|

        Deletes the team.

        Raises
        -------
        Forbidden
            You do not have permissions to delete the team.
        HTTPException
            Deleting the team failed.
        """
        await self._state.http.delete_team(self.id)


class TeamMember(User):
    """Represents a team member in a team.

    .. container:: operations

        .. describe:: x == y

            Checks if two team members are equal.

        .. describe:: x != y

            Checks if two team members are not equal.

        .. describe:: hash(x)

            Return the team member's hash.

        .. describe:: str(x)

            Returns the team member's name with discriminator.

    .. versionadded:: 1.3

    Attributes
    -------------
    team: :class:`Team`
        The team that the member is from.
    membership_state: :class:`ApplicationMembershipState`
        The membership state of the member (i.e. invited or accepted)
    permissions: List[:class:`str`]
        The permissions of the team member. This is always "*".
    """

    __slots__ = ("team", "membership_state", "permissions")

    def __init__(self, team: Team, state: ConnectionState, data: TeamMemberPayload):
        self.team: Team = team
        self.membership_state: ApplicationMembershipState = try_enum(
            ApplicationMembershipState, data["membership_state"]
        )
        self.permissions: List[str] = data.get("permissions", ["*"])
        super().__init__(state=state, data=data["user"])

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} name={self.name!r} "
            f"discriminator={self.discriminator!r} membership_state={self.membership_state!r}>"
        )

    async def remove(self) -> None:
        """|coro|

        Removes the member from the team.

        Raises
        -------
        Forbidden
            You do not have permissions to remove the member.
        HTTPException
            Removing the member failed.
        """
        await self._state.http.remove_team_member(self.team.id, self.id)


class TeamPayout(Hashable):
    """Represents a team payout.

    .. container:: operations

        .. describe:: x == y

            Checks if two team payouts are equal.

        .. describe:: x != y

            Checks if two team payouts are not equal.

        .. describe:: hash(x)

            Return the team payout's hash.

    .. versionadded:: 2.0

    Attributes
    -----------
    id: :class:`int`
        The ID of the payout.
    user_id: :class:`int`
        The ID of the user who is to be receiving the payout.
    status: :class:`PayoutStatus`
        The status of the payout.
    amount: :class:`int`
        The amount of the payout.
    period_start: :class:`datetime.date`
        The start of the payout period.
    period_end: Optional[:class:`datetime.date`]
        The end of the payout period, if ended.
    payout_date: Optional[:class:`datetime.date`]
        The date the payout was made, if made.
    tipalti_submission_response: Optional[:class:`Metadata`]
        The latest response from Tipalti, if exists.
    """

    def __init__(self, *, data: TeamPayoutPayload, team: Team):
        self.team: Team = team

        self.id: int = int(data["id"])
        self.user_id: int = int(data["user_id"])
        self.status: PayoutStatus = try_enum(PayoutStatus, data["status"])
        self.amount: int = data["amount"]
        self.period_start: date = utils.parse_date(data["period_start"])
        self.period_end: Optional[date] = utils.parse_date(data.get("period_end"))
        self.payout_date: Optional[date] = utils.parse_date(data.get("payout_date"))
        self.tipalti_submission_response: Optional[Metadata] = (
            Metadata(data["latest_tipalti_submission_response"])
            if "latest_tipalti_submission_response" in data
            else None
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} status={self.status!r}>"

    async def report(self, type: PayoutReportType) -> bytes:
        """|coro|

        Returns the report for the payout in CSV format.

        Parameters
        -----------
        type: :class:`PayoutReportType`
            The type of report to get the URL for.

        Raises
        -------
        Forbidden
            You do not have permissions to get the report URL.
        HTTPException
            Getting the report URL failed.

        Returns
        -------
        :class:`bytes`
            The report content.
        """
        return await self.team._state.http.get_team_payout_report(
            self.team.id, self.id, str(type)
        )
